/*
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *   * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 */

package org.wso2.registry.secure;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.authenticator.Authenticator;
import org.wso2.authenticator.AuthenticatorException;
import org.wso2.registry.*;
import org.wso2.registry.exceptions.ResourceNotFoundException;
import org.wso2.registry.users.Authorizer;
import org.wso2.registry.users.UserRealm;
import org.wso2.registry.users.UserStoreException;
import org.wso2.registry.users.accesscontrol.AuthorizingRealm;
import org.wso2.registry.users.accesscontrol.AuthorizingRealmConfig;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Map;

/**
 * Registry implementation to perform authentication and authorization. Each
 * user should use a separate instance of the SecureRegistry initialized with
 * the user name and a back-end Registry. All SecureRegistry instances should be
 * initialized using the same back-end registry implementation. SecureRegistry
 * verifies whether the user has necessary permissions to perform the requested
 * action and delegates the request to the back-end registry if the
 * authorization succeeds. It also sets user name as a thread local variable
 * before delegating the request, so that the back-end registry can use that
 * user name to associate user information with performed operation (e.g. who
 * commented on the resource /projects/abcbank/orders.doc). <p/> Resource path
 * is used as the unique ID of the resource to associate permissions. So when an
 * operation is request on a certain path, we can first check the permissions
 * for that path, without even touching the actual resource. This has a side
 * effect that if a user has requested an operation on a non-existing resource,
 * he gets an unauthorized operation message rather than a resource does not
 * exist message. This is good as we don't want unauthorized users to check if a
 * resource exist in a particular path.
 */
public class SecureRegistry implements Registry {

    private static final Log log = LogFactory.getLog(SecureRegistry.class);

    private String userID = RegistryConstants.ANONYMOUS_USER;
    private Registry registry;
    private Authorizer authorizer;
    private UserRealm userRealm;

    // private RegistryUserManager userManager;

    public SecureRegistry(String userID, String password, Registry registry,
            UserRealm defaultRealm) throws RegistryException {

        UserRealm userRealm;
        try {
            userRealm = new AuthorizingRealm();
            AuthorizingRealmConfig config = (AuthorizingRealmConfig) userRealm
                    .getRealmConfiguration();
            config.setRealm(defaultRealm);
            config.setAuthenticatedUserName(userID);
            config.setCurrentUserReadable(true);
            config.setCurrentUserEditable(true);
            userRealm.init(config);
            this.userRealm = userRealm;

            Authenticator authenticator = userRealm.getAuthenticator();
            authorizer = userRealm.getAuthorizer();

            if (!authenticator.authenticate(userID, password)) {
                String msg = "Attempted to authenticate invalid user.";
                log.info(msg);
                throw new RegistryException(msg);
            }
        } catch (AuthenticatorException e){
            String msg = "Could not check authentication. \nCaused by "
                    + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        } catch (UserStoreException e){
            String msg = "Could not check authentication. \nCaused by "
                + e.getMessage();
            log.error(msg, e);
        throw new RegistryException(msg, e);
        }

        this.userID = userID;
        this.registry = registry;
        // this.userManager = new RegistryUserManager(userRealm);
    }

    /**
     * Non-authenticating constructor. Use this constructor if the user is
     * already authenticated outside the secure registry.
     * 
     * @param userID
     *            user name of the user of the secure registry.
     * @param registry
     *            core registry to use by the secure registry.
     * @param defaultRealm
     *            default (non-authorizing) realm.
     * @throws RegistryException
     */
    public SecureRegistry(String userID, Registry registry, UserRealm defaultRealm)
            throws RegistryException {

        UserRealm userRealm;
        try {

            userRealm = new AuthorizingRealm();
            AuthorizingRealmConfig config = (AuthorizingRealmConfig) userRealm
                    .getRealmConfiguration();
            config.setRealm(defaultRealm);
            config.setAuthenticatedUserName(userID);
            userRealm.init(config);
            this.userRealm = userRealm;

            authorizer = userRealm.getAuthorizer();

        } catch (UserStoreException e) {
            String msg = "Could not check authentication. \nCaused by "
                    + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        this.userID = userID;
        this.registry = registry;
        // this.userManager = new RegistryUserManager(userRealm);
    }

    public String getUserID() {
        return userID;
    }

    public UserRealm getUserRealm() {
        return userRealm;
    }

    public Resource get(String path) throws RegistryException {

        if (!path.contains(RegistryConstants.URL_SEPARATOR)
                && !path.contains("?") && !registry.resourceExists(path)) {
            throw new ResourceNotFoundException(path);
        }

        // if the user has permission to the current version, he will get
        // permission to all
        // previous versions of the same resource
        String authorizationPath = getAuthorizationPath(path);

        try {
            if (!authorizer.isUserAuthorized(userID, authorizationPath,
                    ActionConstants.GET)) {
                String msg = "Attempted to perform unauthorized operation ("
                        + ActionConstants.GET + ") on " + authorizationPath
                        + " by " + userID;
                log.info(msg);
                throw new AuthorizationFailedException(msg);
            }
        } catch (UserStoreException e) {
            String msg = "Could not check authorization. \nCaused by "
                    + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        User.setCurrentUser(userID);
        return registry.get(path);
    }

    public boolean resourceExists(String resourcePath) throws RegistryException {

        User.setCurrentUser(userID);
        return registry.resourceExists(resourcePath);
    }

    /**
     * Checks if the user has "PUT" permissions to the existing ascendent path
     * of the given path. e.g. if a user wants to put a resource in to the path
     * "/d1/d2/d3/r1" and only "/d1/d2" currently exist in the registry, that
     * user should have PUT permission to "/d1/d2". <p/> If the user has PUT
     * permission, delegates the request to the underlying jdbc registry. <p/>
     * Then sets the default permissions for newly created paths. i.e. in the
     * above sample, default permissions are set for "/d1/d2/d3" and
     * "/d1/d2/d3/r1".
     * 
     * @param suggestedPath
     *            the path which we'd like to use for the new resource.
     * @param resource
     *            New resource
     * @throws RegistryException
     */
    public String put(String suggestedPath, Resource resource)
            throws RegistryException {

        // find the existing ascendent path of the given path
        isUuserAllowToAddResource(suggestedPath);

        User.setCurrentUser(userID);

        // boolean updatedOnly = false;
        // if (registry.resourceExists(path)) {
        // updatedOnly = true;
        // }

        registry.put(suggestedPath, resource);

        // we should set default permissions for all newly created paths
        // if (!updatedOnly) {
        // String newPath =
        // RegistryConstants.ROOT_PATH.equals(existingAscendant)?
        // "": existingAscendant;
        // for (int i = partCount + 1; i < parts.length; i++) {
        // newPath = newPath + RegistryConstants.PATH_SEPARATOR + parts[i];
        // setDefaultAuthorizations(newPath, userID);
        // }
        // }
        return suggestedPath;
    }

    private void isUuserAllowToAddResource(String suggestedPath)
            throws RegistryException {
        String pathToSplit = suggestedPath;
        if (pathToSplit.startsWith(RegistryConstants.PATH_SEPARATOR)) {
            pathToSplit = pathToSplit
                    .substring(RegistryConstants.PATH_SEPARATOR.length());
        }

        if (pathToSplit.endsWith(RegistryConstants.PATH_SEPARATOR)) {
            pathToSplit = pathToSplit.substring(0, pathToSplit.length()
                    - RegistryConstants.PATH_SEPARATOR.length());
        }

        String[] parts = pathToSplit.split(RegistryConstants.PATH_SEPARATOR);
        int partCount = parts.length;
        String existingAscendant;
        do {
            existingAscendant = "";
            if (partCount == 0) {
                existingAscendant = RegistryConstants.ROOT_PATH;
            }

            for (int i = 0; i < partCount; i++) {
                existingAscendant = existingAscendant
                        + RegistryConstants.PATH_SEPARATOR + parts[i];
            }
            partCount--;
        } while (!registry.resourceExists(existingAscendant));

        // check if the user has put permission for that path
        try {
            if (!authorizer.isUserAuthorized(userID, existingAscendant,
                    ActionConstants.PUT)) {
                String msg = "Attempted to perform unauthorized operation.";
                log.info(msg);
                throw new AuthorizationFailedException(msg);
            }
        } catch (UserStoreException e) {
            String msg = "Could not check authorization. \nCaused by "
                    + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public String importResource(String suggestedPath, String sourceURL,
            Resource metadata) throws RegistryException {

        // find the existing ascendent path of the given path
        isUuserAllowToAddResource(suggestedPath);

        User.setCurrentUser(userID);

        registry.importResource(suggestedPath, sourceURL, metadata);

        return suggestedPath;
    }

    public void delete(String path) throws RegistryException {
        String authPath = getAuthorizationPath(path);
        try {
            if (path.indexOf(";") == -1) {

                // if the request is for deleting an actual resource, we should
                // check if the
                // user has delete permission of the resource. If the request is
                // for deleting
                // a resource metadata, the permission stuff should also be
                // handled in
                // the url handlers. Because there is no way to interpret the
                // path and identify
                // the action at his point.

                if (!authorizer.isUserAuthorized(userID, authPath,
                        ActionConstants.DELETE)) {
                    String msg = "Attempted to perform unauthorized operation.";
                    log.info(msg);
                    throw new AuthorizationFailedException(msg);
                }

            } else {

                // user is trying to perform a delete on resource metadata. so
                // put permission
                // is sufficient.
                // if (!authorizer.isUserAuthorized(userID, authPath,
                // ActionConstants.PUT)) {
                // String msg = "Attempted to perform unauthorized operation.";
                // log.info(msg);
                // throw new AuthorizationFailedException(msg);
                // }
            }
        } catch (UserStoreException e) {
            String msg = "Could not check authorization. \nCaused by "
                    + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        User.setCurrentUser(userID);
        registry.delete(path);
    }

    public String rename(String currentPath, String newPath)
            throws RegistryException {
        if (resourceExists(currentPath)) {
            try {

                String authPath = getAuthorizationPath(currentPath);
                if (!authorizer.isUserAuthorized(userID, authPath,
                        ActionConstants.DELETE)) {
                    String msg = "Attempted to perform unauthorized operation.";
                    log.info(msg);
                    throw new AuthorizationFailedException(msg);
                }
            } catch (UserStoreException e) {
                String msg = "Could not check authorization. \nCaused by "
                        + e.getMessage();
                log.error(msg, e);
                throw new RegistryException(msg);
            }
            isUuserAllowToAddResource(newPath);
            User.setCurrentUser(userID);
            return registry.rename(currentPath, newPath);
        } else {
            throw new ResourceNotFoundException(currentPath);
        }
    }

    public String[] getVersions(String resourcePath) throws RegistryException {

        String authorizationPath = getAuthorizationPath(resourcePath);

        try {
            if (!authorizer.isUserAuthorized(userID, authorizationPath,
                    ActionConstants.GET)) {
                String msg = "User: "
                        + userID
                        + " is not authorized to get versions of the resource: "
                        + authorizationPath;
                log.info(msg);
                throw new AuthorizationFailedException(msg);
            }
        } catch (UserStoreException e) {
            String msg = "Could not check authorization. \nCaused by "
                    + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        User.setCurrentUser(userID);
        return registry.getVersions(resourcePath);
    }

    public void restoreVersion(String versionPath) throws RegistryException {

        String authorizationPath = getAuthorizationPath(versionPath);

        try {
            if (!authorizer.isUserAuthorized(userID, authorizationPath,
                    ActionConstants.PUT)) {
                String msg = "User: " + userID
                        + " is not authorized to restore the resource: "
                        + authorizationPath;
                log.info(msg);
                throw new AuthorizationFailedException(msg);
            }
        } catch (UserStoreException e) {
            String msg = "Could not check authorization. \nCaused by "
                    + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        User.setCurrentUser(userID);
        registry.restoreVersion(versionPath);
    }

    /**
     * Checks if the current user has PUT permission for the dependentPath. User should have write
     * permission to specify a resourceis dependent on another resource.
     *
     * @param dependentPath Path of the dependent resource
     * @param dependencyPaths
     * @throws RegistryException On authorizing error or an error caused by the underlying core
     * registry impl.
     */
    public void addDependencies(String dependentPath, String[] dependencyPaths) throws RegistryException {

        String authPath = getAuthorizationPath(dependentPath);

        try {
            if (!authorizer.isUserAuthorized(userID, authPath, ActionConstants.PUT)) {
                String msg = "User: " + userID + " does not have write permission on the resource: " +
                             authPath + ". Users should have write permission to specify that " +
                             "resource is dependent on another resource.";
                log.info(msg);
                throw new AuthorizationFailedException(msg);
            }
        } catch (UserStoreException e) {
            String msg = "Could not check authorization. \nCaused by " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        registry.addDependencies(dependentPath, dependencyPaths);
    }

    public DependencyChain[] getAllDependencies(String resourcePath) throws RegistryException {

        // todo: stop the dependency chain whenever a resource without the read permission is encountered.
        
        return registry.getAllDependencies(resourcePath);
    }

    public DependentChain[] getAllDependents(String resourcePath) throws RegistryException {

        // todo: stop the dependent chain whenever a resource without the read permission is encountered.

        return registry.getAllDependents(resourcePath);
    }

    public void applyTag(String resourcePath, String tag) throws RegistryException {

        String authorizationPath = getAuthorizationPath(resourcePath);

        try {
            if (!authorizer.isUserAuthorized(userID, authorizationPath,
                    ActionConstants.GET)) {
                String msg = "User: " + userID
                        + " is not authorized to tag on the resource: "
                        + authorizationPath;
                log.info(msg);
                throw new AuthorizationFailedException(msg);
            }
        } catch (UserStoreException e) {
            String msg = "Could not check authorization. \nCaused by "
                    + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        User.setCurrentUser(userID);
        registry.applyTag(resourcePath, tag);
    }

    public TaggedResourcePath[] getResourcePathsWithTag(String tag)
            throws RegistryException {
        User.setCurrentUser(userID);
        return registry.getResourcePathsWithTag(tag);
    }

    public Tag[] getTags(String resourcePath) throws RegistryException {

        String authorizationPath = getAuthorizationPath(resourcePath);

        try {
            if (!authorizer.isUserAuthorized(userID, authorizationPath,
                    ActionConstants.GET)) {
                String msg = "User: " + userID
                        + " is not authorized to get tags of the resource: "
                        + authorizationPath;
                log.info(msg);
                throw new AuthorizationFailedException(msg);
            }
        } catch (UserStoreException e) {
            String msg = "Could not check authorization. \nCaused by "
                    + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        User.setCurrentUser(userID);
        return registry.getTags(resourcePath);
    }

    public void removeTag(String resourcePath, String tag)
            throws RegistryException {

        String authorizationPath = getAuthorizationPath(resourcePath);

        try {
            if (!authorizer.isUserAuthorized(userID, authorizationPath,
                    ActionConstants.GET)) {
                String msg = "User: " + userID
                        + " is not authorized to remove tags of the resource: "
                        + authorizationPath;
                log.info(msg);
                throw new AuthorizationFailedException(msg);
            }
        } catch (UserStoreException e) {
            String msg = "Could not check authorization. \nCaused by "
                    + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        User.setCurrentUser(userID);
        registry.removeTag(resourcePath, tag);
    }

    public String addComment(String resourcePath, Comment comment)
            throws RegistryException {

        String authorizationPath = getAuthorizationPath(resourcePath);

        try {
            if (!authorizer.isUserAuthorized(userID, authorizationPath,
                    ActionConstants.GET)) {
                String msg = "User: " + userID
                        + " is not authorized to comment on the resource: "
                        + authorizationPath;
                log.info(msg);
                throw new AuthorizationFailedException(msg);
            }
        } catch (UserStoreException e) {
            String msg = "Could not check authorization. \nCaused by "
                    + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        User.setCurrentUser(userID);
        return registry.addComment(resourcePath, comment);
    }

    public void editComment(String commentPath, String text)
            throws RegistryException {

        String authorizationPath = getAuthorizationPath(commentPath);

        try {
            if (!authorizer.isUserAuthorized(userID, authorizationPath,
                    ActionConstants.GET)) {
                String msg = "User: " + userID
                        + " is not authorized to comment on the resource: "
                        + authorizationPath;
                log.info(msg);
                throw new AuthorizationFailedException(msg);
            }
        } catch (UserStoreException e) {
            String msg = "Could not check authorization. \nCaused by "
                    + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        User.setCurrentUser(userID);
        registry.editComment(commentPath, text);
    }

    public Comment[] getComments(String resourcePath) throws RegistryException {
        User.setCurrentUser(userID);
        return registry.getComments(resourcePath);
    }

    public void rateResource(String resourcePath, int rating)
            throws RegistryException {

        String authorizationPath = getAuthorizationPath(resourcePath);

        try {
            if (!authorizer.isUserAuthorized(userID, authorizationPath,
                    ActionConstants.GET)) {
                String msg = "User: " + userID
                        + " is not authorized to rate the resource: "
                        + authorizationPath;
                log.info(msg);
                throw new AuthorizationFailedException(msg);
            }
        } catch (UserStoreException e) {
            String msg = "Could not check authorization. \nCaused by "
                    + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        User.setCurrentUser(userID);
        registry.rateResource(resourcePath, rating);
    }

    public float getAverageRating(String resourcePath) throws RegistryException {

        String authorizationPath = getAuthorizationPath(resourcePath);

        try {
            if (!authorizer.isUserAuthorized(userID, authorizationPath,
                    ActionConstants.GET)) {
                String msg = "User: "
                        + userID
                        + " is not authorized to get average rating of the resource: "
                        + authorizationPath;
                log.info(msg);
                throw new AuthorizationFailedException(msg);
            }
        } catch (UserStoreException e) {
            String msg = "Could not check authorization. \nCaused by "
                    + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        User.setCurrentUser(userID);
        return registry.getAverageRating(resourcePath);
    }

    public int getRating(String resourcePath, String userName)
            throws RegistryException {

        String authorizationPath = getAuthorizationPath(resourcePath);

        try {
            if (!authorizer.isUserAuthorized(userID, authorizationPath,
                    ActionConstants.GET)) {
                String msg = "User: " + userID
                        + " is not authorized to get ratings of the resource: "
                        + authorizationPath;
                log.info(msg);
                throw new AuthorizationFailedException(msg);
            }
        } catch (UserStoreException e) {
            String msg = "Could not check authorization. \nCaused by "
                    + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        User.setCurrentUser(userID);
        return registry.getRating(resourcePath, userName);
    }

    public Resource executeQuery(String path, Map parameters)
            throws RegistryException {
        User.setCurrentUser(userID);
        return registry.executeQuery(path, parameters);
    }

    public LogEntry[] getLogs(String resourcePath, int action, String userName,
            Date from, Date to, boolean recentFirst) throws RegistryException {

        User.setCurrentUser(userID);

        LogEntry[] allEntries = registry.getLogs(resourcePath, action,
                userName, from, to, recentFirst);

        // remove logs about resource for which the current user doesn't get
        // read permission.
        List authorizedEnListList = new ArrayList();
        for (LogEntry allEntry : allEntries) {
            try {
                if (authorizer.isUserAuthorized(userID, allEntry
                        .getResourcePath(), ActionConstants.GET)) {
                    authorizedEnListList.add(allEntry);
                }
            } catch (UserStoreException e) {
                log.error("Could not check permissions for resource: "
                        + allEntry.getResourcePath(), e);
            }
        }

        return (LogEntry[]) authorizedEnListList
                .toArray(new LogEntry[authorizedEnListList.size()]);
    }

    /**
     * Path of a resource given to the Registry interface may contain extensions
     * to refer metadata about resources. But we always store the authorization
     * for resources against the pure resource path, stored in the ARTIFACTS
     * table. This methods extracts that pure resource path from a given path.
     * 
     * @param resourcePath
     *            A path string, which may contain extensions
     * @return pure resource path for the given path
     */
    private String getAuthorizationPath(String resourcePath) {

        // if the user has permission to the current version, he will get
        // permission to all
        // previous versions of the same resource

        String preparedPath = resourcePath;
        if (resourcePath.indexOf("?") > 0) {
            preparedPath = resourcePath.split("\\?")[0];
        } else if (resourcePath.indexOf(";") > 0) {
            preparedPath = resourcePath.split("\\;")[0];
        }

        if (preparedPath.equals(RegistryConstants.ROOT_PATH)) {
            return preparedPath;

        } else {

            if (!preparedPath.startsWith(RegistryConstants.ROOT_PATH)) {
                preparedPath = RegistryConstants.ROOT_PATH + preparedPath;
            }

            if (preparedPath.endsWith(RegistryConstants.PATH_SEPARATOR)) {
                preparedPath = preparedPath.substring(0,
                        preparedPath.length() - 1);
            }
        }

        return preparedPath;
    }
}
